From: Masami Hiramatsu <mhira...@kernel.org>

[ Upstream commit 5bb4fc2d8641219732eb2bb654206775a4219aca ]

Disable preemption in ftrace-based jprobe handlers as
described in Documentation/kprobes.txt:

  "Probe handlers are run with preemption disabled."

This will fix jprobes behavior when CONFIG_PREEMPT=y.

Signed-off-by: Masami Hiramatsu <mhira...@kernel.org>
Cc: Alexei Starovoitov <a...@fb.com>
Cc: Alexei Starovoitov <a...@kernel.org>
Cc: Ananth N Mavinakayanahalli <ana...@linux.vnet.ibm.com>
Cc: Linus Torvalds <torva...@linux-foundation.org>
Cc: Paul E . McKenney <paul...@linux.vnet.ibm.com>
Cc: Peter Zijlstra <pet...@infradead.org>
Cc: Steven Rostedt <rost...@goodmis.org>
Cc: Thomas Gleixner <t...@linutronix.de>
Link: 
http://lkml.kernel.org/r/150581530024.32348.9863783558598926771.stgit@devbox
Signed-off-by: Ingo Molnar <mi...@kernel.org>
Signed-off-by: Sasha Levin <alexander.le...@verizon.com>
---
 arch/x86/kernel/kprobes/ftrace.c | 23 ++++++++++++++---------
 1 file changed, 14 insertions(+), 9 deletions(-)

diff --git a/arch/x86/kernel/kprobes/ftrace.c b/arch/x86/kernel/kprobes/ftrace.c
index 5f8f0b3cc674..2c0b0b645a74 100644
--- a/arch/x86/kernel/kprobes/ftrace.c
+++ b/arch/x86/kernel/kprobes/ftrace.c
@@ -26,7 +26,7 @@
 #include "common.h"
 
 static nokprobe_inline
-int __skip_singlestep(struct kprobe *p, struct pt_regs *regs,
+void __skip_singlestep(struct kprobe *p, struct pt_regs *regs,
                      struct kprobe_ctlblk *kcb, unsigned long orig_ip)
 {
        /*
@@ -41,20 +41,21 @@ int __skip_singlestep(struct kprobe *p, struct pt_regs 
*regs,
        __this_cpu_write(current_kprobe, NULL);
        if (orig_ip)
                regs->ip = orig_ip;
-       return 1;
 }
 
 int skip_singlestep(struct kprobe *p, struct pt_regs *regs,
                    struct kprobe_ctlblk *kcb)
 {
-       if (kprobe_ftrace(p))
-               return __skip_singlestep(p, regs, kcb, 0);
-       else
-               return 0;
+       if (kprobe_ftrace(p)) {
+               __skip_singlestep(p, regs, kcb, 0);
+               preempt_enable_no_resched();
+               return 1;
+       }
+       return 0;
 }
 NOKPROBE_SYMBOL(skip_singlestep);
 
-/* Ftrace callback handler for kprobes */
+/* Ftrace callback handler for kprobes -- called under preepmt disabed */
 void kprobe_ftrace_handler(unsigned long ip, unsigned long parent_ip,
                           struct ftrace_ops *ops, struct pt_regs *regs)
 {
@@ -77,13 +78,17 @@ void kprobe_ftrace_handler(unsigned long ip, unsigned long 
parent_ip,
                /* Kprobe handler expects regs->ip = ip + 1 as breakpoint hit */
                regs->ip = ip + sizeof(kprobe_opcode_t);
 
+               /* To emulate trap based kprobes, preempt_disable here */
+               preempt_disable();
                __this_cpu_write(current_kprobe, p);
                kcb->kprobe_status = KPROBE_HIT_ACTIVE;
-               if (!p->pre_handler || !p->pre_handler(p, regs))
+               if (!p->pre_handler || !p->pre_handler(p, regs)) {
                        __skip_singlestep(p, regs, kcb, orig_ip);
+                       preempt_enable_no_resched();
+               }
                /*
                 * If pre_handler returns !0, it sets regs->ip and
                 * resets current kprobe.
                 */
        }
 end:
@@ -77,13 +78,17 @@ void kprobe_ftrace_handler(unsigned long ip, unsigned long 
                /*
                 * If pre_handler returns !0, it sets regs->ip and
-                * resets current kprobe.
+                * resets current kprobe, and keep preempt count +1.
                 */
        }
-- 
2.11.0

Reply via email to